"""
PyTorch dataset wrapper for the
`Adult (Census Income) <https://archive.ics.uci.edu/dataset/2/adult>`_ dataset.
"""
__version__ = "0.1.0"

from .adult import Adult
from .default import Default
from .law_school import LawSchool
from .south_german import SouthGerman
